SystemOrganization addCategory: #GODProcessors!

Object subclass: #PCSGODTagger
	instanceVariableNames: 'tagger'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODProcessors'!

!PCSGODTagger methodsFor: 'tagging' stamp: 'TD 11/16/2014 17:36'!
addTagsTo: aGodData
 
	|tags|

	" Get most relevants tokens in aGodData and add it as tags of aGodData "
	
	tags := (tagger getMoreRelevantsOf: aGodData content) asOrderedCollection.
	aGodData tags: tags.! !

!PCSGODTagger methodsFor: 'initialize-release' stamp: 'TD 11/16/2014 17:35'!
initialize

	"define a standard configuration to the algorithm"
	tagger := PCSTagger new.! !

!PCSGODTagger methodsFor: 'tagging' stamp: 'TD 11/16/2014 17:49'!
tagCollection: aGODDataCollection

	self training: aGODDataCollection.
	aGODDataCollection do: [ :element |
		self addTagsTo: element
  	].! !

!PCSGODTagger methodsFor: 'accessing' stamp: 'TD 11/16/2014 17:52'!
tagger
	"get tagger"
	^tagger.! !

!PCSGODTagger methodsFor: 'accessing' stamp: 'TD 11/16/2014 17:52'!
tagger: value
	"set tagger"
	tagger := value.! !

!PCSGODTagger methodsFor: 'tagging' stamp: 'TD 11/16/2014 17:36'!
training: aGODDataCollection
	|aStringCollection|
	
  	"it extracts the aGODDataCollection's contents"
	aStringCollection := Bag new.
  	aGODDataCollection do: [ :aElement |
		aStringCollection add: aElement content.
    	].
  	"it trains the algorithm with the given aGODDataCollection contents"
	tagger createDictIdf: aStringCollection.! !

Object subclass: #PCSGODTaggerExample
	instanceVariableNames: 'tagger aGODCollection godTagger'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODProcessors'!

!PCSGODTaggerExample methodsFor: 'initialize-release' stamp: 'TD 11/23/2014 18:20'!
initialize
	| p |
	"Config tagger and preprocessor"
	godTagger := PCSGODTagger new.
	tagger := PCSTagger new.
	p := PCSEnglishPreprocessor new.
	"using preprocessor fileType as HTML is enough to deal with XML files"
	p fileType: 'HTML'.
	tagger preprocessor: p.
	godTagger tagger: tagger.
	
	"Load files to aGODCollection"
	self loadFiles.! !

!PCSGODTaggerExample methodsFor: 'initialize-release' stamp: 'TD 11/23/2014 18:28'!
loadFiles
	| aGODData aString corpusDir dirName ios x |
	dirName := (FileDirectory default pathName),'/corpus/'.
	corpusDir := FileDirectory forFileName: dirName.
	aGODCollection := Bag new.

	corpusDir fileAndDirectoryNames do:[:f | ios := FileStream oldFileNamed: dirName,f.
		aString := ''.
		[(x := ios nextLine) notNil]
		whileTrue: [
			aString := aString,x.
		].
		aGODData := GODData new.
		aGODData content: aString.
		aGODData title: f.
		aGODCollection add: aGODData.
		Transcript show: f, ' loaded'; cr.
	].
	Transcript show: aGODCollection size, ' files loaded'; cr;cr.! !

!PCSGODTaggerExample methodsFor: 'run' stamp: 'TD 11/23/2014 18:45'!
run
	"traine with aGODCollection"
	godTagger training: aGODCollection.

	"add tags do elements of aGODData"
	aGODCollection do:[ :aGODData |
		godTagger addTagsTo: aGODData.
	].

	aGODCollection do:[ :aGODData |
		Transcript show: aGODData title, ' -> '; show: aGODData tags;cr.	
	].! !

Object subclass: #PCSPreprocessor
	instanceVariableNames: 'fileType punctuation removeStopWords languageStopWords stopWords'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODProcessors'!

!PCSPreprocessor commentStamp: 'TD 11/16/2014 17:12' prior: 0!
A PCSPreprocessor is xxxxxxxxx.

Instance Variables
	fileType:	<String>
		

fileType
	-Defines the type of arquive that is treated
	May be one of two types:
		-'TXT': a commom file 
		-'HTML': a HTML file

!

PCSPreprocessor subclass: #PCSEnglishPreprocessor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODProcessors'!

!PCSEnglishPreprocessor methodsFor: 'initialize-release' stamp: 'TD 11/20/2014 15:01'!
initialize
	"define a standard configuration to the stopwords in English"
	super initialize.
	
	"stopwrods from www.ranks.nl/sopwords"
	self stopWords: #('a' 'about' 'above' 'after' 'again' 'against' 'all' 'am' 'an' 'and' 'any' 'are' 'aren''t' 'as' 'at' 'be' 'because' 'been' 'before' 'being' 'below' 'between' 'both' 'but' 'by' 'can''t' 'cannot' 'could' 'couldn''t' 'did' 'didn''t' 'do' 'does' 'doesn''t' 'doing' 'don''t' 'down' 'during' 'each' 'few' 'for' 'from' 'further' 'had' 'hadn''t' 'has' 'hasn''t' 'have' 'haven''t' 'having' 'he' 'he''d' 'he''ll' 'he''s' 'her' 'here' 'here''s' 'hers' 'herself' 'him' 'himself' 'his' 'how' 'how''s' 'i' 'i''d' 'i''ll' 'i''m' 'i''ve' 'if' 'in' 'into' 'is' 'isn''t' 'it' 'it''s' 'its' 'itself' 'let''s' 'me' 'more' 'most' 'mustn''t' 'my' 'myself' 'no' 'nor' 'not' 'of' 'off' 'on' 'once' 'only' 'or' 'other' 'ought' 'our' 'ours' 'ourselves' 'out' 'over' 'own' 'same' 'shan''t' 'she' 'she''d' 'she''ll' 'she''s' 'should' 'shouldn''t' 'so' 'some' 'such' 'than' 'that' 'that''s' 'the' 'their' 'theirs' 'them' 'themselves' 'then' 'there' 'there''s' 'these' 'they' 'they''d' 'they''ll' 'they''re' 'they''ve' 'this' 'those' 'through' 'to' 'too' 'under' 'until' 'up' 'very' 'was' 'wasn''t' 'we' 'we''d' 'we''ll' 'we''re' 'we''ve' 'were' 'weren''t' 'what' 'what''s' 'when' 'when''s' 'where' 'where''s' 'which' 'while' 'who' 'who''s' 'whom' 'why' 'why''s' 'with' 'won''t' 'would' 'wouldn''t' 'you' 'you''d' 'you''ll' 'you''re' 'you''ve' 'your' 'yours' 'yourself' 'yourselves') asSet. ! !

PCSPreprocessor subclass: #PCSPortuguesePreprocessor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODProcessors'!

!PCSPortuguesePreprocessor methodsFor: 'initialize-release' stamp: 'TD 11/20/2014 15:04'!
initialize
	"define a standard configuration to the stopwords in Portuguese"
	
	super initialize.
	"stopwrods from www.ranks.nl/sopwords/portuguese"
	self stopWords: #('ltimo' '' 'acerca' 'agora' 'algmas' 'alguns' 'ali' 'ambos' 'antes' 'apontar' 'aquela' 'aquelas' 'aquele' 'aqueles' 'aqui' 'atrs' 'bem' 'bom' 'cada' 'caminho' 'cima' 'com' 'como' 'comprido' 'conhecido' 'corrente' 'das' 'debaixo' 'dentro' 'desde' 'desligado' 'deve' 'devem' 'dever' 'direita' 'diz' 'dizer' 'dois' 'dos' 'e' 'ela' 'ele' 'eles' 'em' 'enquanto' 'ento' 'est' 'esto' 'estado' 'estar' 'estar' 'este' 'estes' 'esteve' 'estive' 'estivemos' 'estiveram' 'eu' 'far' 'faz' 'fazer' 'fazia' 'fez' 'fim' 'foi' 'fora' 'horas' 'iniciar' 'inicio' 'ir' 'ir' 'ista' 'iste' 'isto' 'ligado' 'maioria' 'maiorias' 'mais' 'mas' 'mesmo' 'meu' 'muito' 'muitos' 'ns' 'no' 'nome' 'nosso' 'novo' 'o' 'onde' 'os' 'ou' 'outro' 'para' 'parte' 'pegar' 'pelo' 'pessoas' 'pode' 'poder' 'podia' 'por' 'porque' 'povo' 'promeiro' 'qu' 'qual' 'qualquer' 'quando' 'quem' 'quieto' 'so' 'saber' 'sem' 'ser' 'seu' 'somente' 'tm' 'tal' 'tambm' 'tem' 'tempo' 'tenho' 'tentar' 'tentaram' 'tente' 'tentei' 'teu' 'teve' 'tipo' 'tive' 'todos' 'trabalhar' 'trabalho' 'tu' 'um' 'uma' 'umas' 'uns' 'usa' 'usar' 'valor' 'veja' 'ver' 'verdade' 'verdadeiro' 'voc') asSet.! !

!PCSPreprocessor methodsFor: 'accessing' stamp: 'TD 11/16/2014 11:27'!
fileType
	"get fileType"
	^fileType. ! !

!PCSPreprocessor methodsFor: 'accessing' stamp: 'TD 11/16/2014 11:27'!
fileType: value
	"set fileType"
	fileType := value.! !

!PCSPreprocessor methodsFor: 'initialize-release' stamp: 'TD 11/20/2014 14:07'!
initialize

	"define a standard configuration to the preprocessor"
	fileType := 'TXT'.
	punctuation := '?!!.,:;()[]{}%$#+='.
	"The class initialize doesn't have a language so it doesn't have any stopWords"
	stopWords := ''.! !

!PCSPreprocessor methodsFor: 'preprocessing' stamp: 'TD 11/20/2014 14:03'!
preprocess: aString
	| aBag treatedString|
  "this method realizes comom preprocess methods over the String
	1- tokenizes the String 
    - may remove common words
    - capitalize tokens
    - may do stemming and lemmatization
	anArray := self tokenizer: aS. 
  "
	treatedString := self treatType: aString. 
	aBag := self tokenizer: treatedString. 
	aBag := self removeStopWordsOf: aBag.
	^aBag.! !

!PCSPreprocessor methodsFor: 'accessing' stamp: 'FKP 11/17/2014 11:40'!
punctuation
	"get punctuation"

	^punctuation.! !

!PCSPreprocessor methodsFor: 'accessing' stamp: 'FKP 11/17/2014 11:43'!
punctuation: aString
	"set punctuation, example '?!!.,:;()[]{}%$#+=' "
	
	punctuation := aString. ! !

!PCSPreprocessor methodsFor: 'preprocessing' stamp: 'TD 11/20/2014 14:01'!
removeStopWordsOf: aBagTokens
	"remove stopwords of aBagTokens"
	
	self stopWords do: [:stopword | aBagTokens do: 
							[:token | stopword = token ifTrue: 
										[aBagTokens remove: token]
							]
						 ].
					
	^aBagTokens.
	
	
	

	! !

!PCSPreprocessor methodsFor: 'accessing' stamp: 'TD 11/20/2014 13:54'!
stopWords
	"get stopWods"

	^stopWords.! !

!PCSPreprocessor methodsFor: 'accessing' stamp: 'TD 11/20/2014 13:55'!
stopWords: aCollectionStopWords
	"set stopWords"

	stopWords := aCollectionStopWords.! !

!PCSPreprocessor methodsFor: 'preprocessing' stamp: 'FKP 11/17/2014 11:43'!
tokenizer: aString

|tokenizedString aBag|

tokenizedString := aString copyWithoutAll: punctuation.
tokenizedString translateToLowercase.
aBag := (tokenizedString substrings) asBag.
^aBag.



! !

!PCSPreprocessor methodsFor: 'preprocessing' stamp: 'TD 11/16/2014 13:31'!
treatType: aString
     " this method deals with different types of string"
	| filter tmpBag newString |
	(fileType = 'TXT') ifTrue: [
		^aString.
		].
	
	(fileType = 'HTML') ifTrue: [
		filter := FILTTextFilter new.
		"Remove HTML tags"
		tmpBag := filter filterText: aString between: '>' and: '<'.
		
		"Recontruct the string, keeping add spaces in the place of the old tags"
		newString := ''.
		tmpBag do:[ :word | newString := newString ,' ', word.].
		^newString.		
		].
	! !

Object subclass: #PCSStatisticsCalculator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODProcessors'!

!PCSStatisticsCalculator commentStamp: 'TD 11/16/2014 17:14' prior: 0!
PCSStatisticsCalculator is a class that calculates statistical measures of central tendency and dispersion, for a given generic collection. 

No selector key, the method applies to the absolute values of the collection and its usage is intuitive.

Already with the selector, the measure is calculated on the key value, for each element of the collection.

The key is a code block.

For instance, if you want the average width as an attribute. 
you can use the command:  
  pcs average: aCollection key: [ :x | x width].

Now, if you want the average occurrences of a word, you can do:
  pcs average: aCollection key: [ :x | x Countin: aword].


Instance Variables: none
!

!PCSStatisticsCalculator methodsFor: 'averages_and_variances' stamp: 'TD 9/27/2014 14:30'!
average: aCollection 
	"compute the average of the collection of real numbers"
	^aCollection average.! !

!PCSStatisticsCalculator methodsFor: 'averages_and_variances' stamp: 'TD 10/7/2014 19:38'!
average: aCollection key: aKey
	"compute the average of key in the collection"
	| aNumberCollection |
	
	aNumberCollection := Bag new: aCollection size.
	
	aCollection do: [ :each | aNumberCollection add: (aKey value: each) ].
	
	^self average: aNumberCollection.! !

!PCSStatisticsCalculator methodsFor: 'counting' stamp: 'TD 11/23/2014 13:24'!
countIn: aString ocurrencesOf: aWord
	"return the number of ocurrences of aWord in aString"
	| n aBag tokenizedString pre|
	"aCollection := Bag new.
	finalCollection := Bag new. 
	aCollection := aString  subStrings. 
	aCollection do: [:word | ((word endsWithAnyOf: #('.' ',' '(' ')' '[' ']' '{' '}' ';' '$' '!!' '?')) & (word size > 1)) ifTrue: [word := word copyFrom: 1 to: (word size -1)].
							word translateToLowercase.
							finalCollection add: word].
	
	n := finalCollection occurrencesOf: (aWord asLowercase).
	^n."
	
	pre := PCSPreprocessor new.
	tokenizedString := aString copyWithoutAll: (pre punctuation).
	tokenizedString translateToLowercase.
	aBag := (tokenizedString substrings) asBag.
	n := aBag occurrencesOf: (aWord asLowercase).
	^n.! !

!PCSStatisticsCalculator methodsFor: 'initializing' stamp: 'TD 9/27/2014 13:35'!
initialize
	! !

!PCSStatisticsCalculator methodsFor: 'averages_and_variances' stamp: 'TD 10/20/2014 14:15'!
median: aCollection 
	"compute the median of the collection of Number"
	^aCollection median.! !

!PCSStatisticsCalculator methodsFor: 'averages_and_variances' stamp: 'TD 10/20/2014 14:15'!
median: aCollection key: aKey
	"compute the median of this key(k) in the collection(c)"
	|aNumberCollection|
	aNumberCollection := Bag new: aCollection size.
	
	aCollection do: [ :each | aNumberCollection add: (aKey value: each) ].
	
	^self median: aNumberCollection.

	! !

!PCSStatisticsCalculator methodsFor: 'averages_and_variances' stamp: 'TD 10/20/2014 14:31'!
std: aCollection 
	"compute the standard deviation of the collection"
	
	^(self var: aCollection) sqrt.! !

!PCSStatisticsCalculator methodsFor: 'averages_and_variances' stamp: 'TD 10/20/2014 14:39'!
std: aCollection key: aKey
	"compute the standard deviation of key in the collection"
	|aNumberCollection|
	aNumberCollection := Bag new: aCollection size.
	
	aCollection do: [ :each | aNumberCollection add: (aKey value: each) ].
	
	^self std: aNumberCollection.! !

!PCSStatisticsCalculator methodsFor: 'averages_and_variances' stamp: 'TD 10/20/2014 14:42'!
var: aCollection 
	"compute the variance of the collection"
	|average sum|
	
	average := self average: aCollection.
	sum := 0.
	aCollection do: [:aNumber| sum := sum + ((aNumber - average) raisedTo: 2 ) ].
	
	^sum / aCollection size.! !

!PCSStatisticsCalculator methodsFor: 'averages_and_variances' stamp: 'TD 10/20/2014 14:30'!
var: aCollection key: aKey
	"compute the variance of the key in the collection"
	|aNumberCollection|
	aNumberCollection := Bag new: aCollection size.
	
	aCollection do: [ :each | aNumberCollection add: (aKey value: each) ].
	
	^self var: aNumberCollection.! !

Object subclass: #PCSTagger
	instanceVariableNames: 'dictIdf minRelevance maxTf minTf numDocuments preprocessor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODProcessors'!

!PCSTagger commentStamp: 'FKP 11/2/2014 21:31' prior: 0!
A PCSTagger generates tags for GODData using your content. 

To this end, it counts with a variety of methods that enable it: 
do tokenization of strings, calculate the quantities tf-idf, organize the tokens in dictionaries tf-idf, and as the level of significance, generates tags for an element (GODData) or several (several GODData in a collection). 
To generate a tag, tf the respective token must belong to the interval [minTf, maxTf] desired;

The methods training, addTagsTo and tagCollection are specialized to deal with GODData.

How to use

  To tag all element of a collection

      pcsTagger = PCSTagger new.
      pcsTagger tagCollection: aGODDataCollection.

  To tag an element

      pcsTagger = PCSTagger new.
      pcsTagger training: aGODDataCollection.
      pcsTagger addTagsTo: aGODData.

Instance Variables
	
	dictIdf: an dictionary of idf; given a term t it returns the respective idf
	
	maxTf: define the max frequence to be considerated
		   * float in range [0..1]

	minTf: define the min frequence to be considerated
		   * float in range [0..1]
		
	minRelevance: define the min value of relevance that will be accepted as a tag
		   * float in range [0..1]

!

!PCSTagger methodsFor: 'classifying' stamp: 'FKP 11/17/2014 19:47'!
countIn: aBag
	"return a Dictionary with the number of ocurrences of each word in aBag"
	|  aSet aDictionary|

	aSet := Set new.
	aDictionary := Dictionary new. 
	aSet := aBag asSet.
	aSet do: [ :aToken | 
		aDictionary add: (Association key: aToken value: (aBag occurrencesOf: aToken))
	].
	 ^aDictionary! !

!PCSTagger methodsFor: 'classifying' stamp: 'TD 11/16/2014 10:55'!
createDictIdf: aStringCollection
  	|dictDf aBag|
  	numDocuments := aStringCollection size.
  	dictDf := Dictionary new.
	
  	"generates the dictionary of frequencies of each document"
  	aStringCollection do: [ :aElement |
		aBag := preprocessor  preprocess: aElement.
		aBag asSet do: [ :token |
			(dictDf includesKey: token)
				ifTrue:[
					dictDf at: token put: ((dictDf at: token) + 1)
				]
				ifFalse:[
					dictDf at: token put: 1
				].
      		
		]
    	].

  	"converte dict_df into a Inverse Document Frequency dictionary"
  	dictIdf := Dictionary new.
  	dictDf keysAndValuesDo: [ :token :freq |
    		dictIdf at: token put: (( numDocuments / (freq +1)) log)
  	].! !

!PCSTagger methodsFor: 'classifying' stamp: 'FKP 11/17/2014 13:37'!
createDictTf: aString
  	|aBag aDictTf|
	
  	" convert aString to an Bag of tokens"
  	aBag := preprocessor preprocess: aString.

  	" calculates number of ocourrences of each token"
  	aDictTf := self countIn: aBag.

   	^aDictTf! !

!PCSTagger methodsFor: 'accessing' stamp: 'TD 10/30/2014 16:23'!
dictIdf
	"get dictIdf"
	^dictIdf.! !

!PCSTagger methodsFor: 'accessing' stamp: 'TD 10/30/2014 16:27'!
dictIdf: value
	"set dictIdf"
	dictIdf := value.! !

!PCSTagger methodsFor: 'classifying' stamp: 'TD 11/16/2014 10:54'!
getMoreRelevantsOf: aString
 
	|dictTf tfIdfDict tfIdf minTfIdf maxTfIdf aBag tags|
	
	(self dictIdf isEmpty)
	ifTrue: [
		DictIdfMustNotBeEmpty signal
	].
  	dictTf :=  self createDictTf: aString.
	aBag := preprocessor preprocess: aString.
  	"calcula o tfidf de cada termo de aDict"
  	tfIdfDict := Dictionary new.
	
	minTfIdf := 1000.
	maxTfIdf := -1000.
  	dictTf keysAndValuesDo: [ :token :tf |
    		tfIdf := self tfIdf: token withFrequence: tf.
		tfIdfDict at: token put: tfIdf.
		
		"calculates the minTfIdf"
   		 (tfIdf < minTfIdf)
			ifTrue: [
      			minTfIdf := tfIdf.
    			].
		"calculates the maxTfIdf"
   		 (tfIdf > maxTfIdf)
			ifTrue: [
      			maxTfIdf := tfIdf.
    			].  		
		(((tf / aBag size) > self maxTf) | ((tf / aBag size) < self minTf)) 
			ifTrue: [
				dictTf removeKey: token.
			].
			
  	].
	
	"feature scaling, used to bring all values into the range [0,1]"
	 tfIdfDict keysAndValuesDo: [ :token :value |
		tfIdfDict at: token put: ((value - minTfIdf) / (maxTfIdf - minTfIdf)).
  	].
	"calcular minTfIdf e maxTfIdf"
	tags := Bag new.
  	tfIdfDict keysAndValuesDo: [ :token :value |
   		 (value > self minRelevance)
			ifTrue: [
      			tags add: token.
    			]
  	].
  	^tags.! !

!PCSTagger methodsFor: 'initialize-release' stamp: 'TD 11/16/2014 10:52'!
initialize

	"define a standard configuration to the algorithm"
	minRelevance := (0.8).
	minTf := (0.0).
	maxTf := (1.0).
	dictIdf := Dictionary new.
	preprocessor := PCSPreprocessor new.! !

!PCSTagger methodsFor: 'accessing' stamp: 'TD 10/30/2014 16:24'!
maxTf
	"get maxTf"
	^maxTf. ! !

!PCSTagger methodsFor: 'accessing' stamp: 'TD 10/30/2014 16:27'!
maxTf: value
	"set maxTf"
	maxTf := value. ! !

!PCSTagger methodsFor: 'accessing' stamp: 'TD 10/30/2014 16:28'!
minRelevance
	"get minRelevance"
	^minRelevance. ! !

!PCSTagger methodsFor: 'accessing' stamp: 'TD 10/30/2014 16:28'!
minRelevance: value
	"set minRelevance"
	minRelevance := value. ! !

!PCSTagger methodsFor: 'accessing' stamp: 'TD 10/30/2014 16:27'!
minTf
	"get minTf"
	^minTf. ! !

!PCSTagger methodsFor: 'accessing' stamp: 'TD 10/30/2014 16:28'!
minTf: value
	"set minTf"
	minTf := value. ! !

!PCSTagger methodsFor: 'accessing' stamp: 'TD 11/16/2014 10:51'!
preprocessor
	"get preprocessor"
	^preprocessor. ! !

!PCSTagger methodsFor: 'accessing' stamp: 'TD 11/16/2014 10:51'!
preprocessor: value
	"set preprocessor"
	preprocessor := value. ! !

!PCSTagger methodsFor: 'classifying' stamp: 'TD 10/31/2014 12:31'!
tfIdf: aTerm withFrequence: aFrequence
	|idf|
	(dictIdf includesKey: aTerm)
		ifTrue: [	idf:= dictIdf at: aTerm.]
		ifFalse: [idf := ((numDocuments / 1) log). ].
	^(aFrequence * idf).! !
